/*
 * Cgipp.cpp
 *
 * The Cgipp library, by Stig Erik Sand, version 0.3 is a
 * major upgrade from versions 0.1.x and 0.2.x. v0.2 was a
 * more or less patched/more portable/less error frequent/less
 * memory demanding version of the Cgicc library, by Stephen
 * Booth, version 1.0. Cgipp v0.3.x cannot be expected to be
 * backward-compatible, though earlier versions which were
 * backward-compatible can be obtained.
 *  
 * Cgicc is a C++ version of the C library cgic by Thomas
 * Boutell, available from sunsite.unc.edu/boutell/cgic/cgic.html
 *  
 * Cgipp/Cgicc is intended to simplify and speed the
 * development of C++ cgi(common gateway interface)
 * applications by providing a simple, user-friendly set
 * of classes that present full cgi functionality.
 *
 * Version 0.1-0.3 has been tested with gcc 2.7.x and
 * libg++ 2.7.x on Linux 1.2.13, IRIX 5.3, DEC OSF/1,
 * Solaris 2.4, Solaris 2.5, SunOS 4.1.3. It has also
 * been tested with Sun CC on Solaris 2.5, and with
 * SGI C++ for IRIX 6.3. It should compile flawlessly on
 * these platforms. Minor changes needed only in the Makefile
 *
 * $Log: Cgipp.cpp,v $
 * Revision 1.7  1997/09/08 08:47:05  stig
 * updated for v0.3
 *
 * Revision 1.6  1997/03/03 11:11:35  stig
 * Major update
 *
 *
 * $Id: Cgipp.cpp,v 1.7 1997/09/08 08:47:05 stig Exp $
 */


#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <cgi/Cgipp.H>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

DebugTool cgidebug;

void
die( int signal ){
  htmlTool	errDoc( cout );
  char	     	*errorText = "Allotted process time expired.";
  time_t 	Now;
  tm 		*Date;
  char 		S[80];
  
  Now = time( NULL );
  Date = localtime( &Now );
#ifdef HAVE_STRFTIME  
  strftime( S, 80, "%A, %B %d, %Y %I:%M:%S %p", Date );
#else
  sprintf( S, "Warning: lacking strftime()..");
#endif  
  errDoc.printHeader().html().head().title().put( "Cgipp Error: " );
  errDoc.put( errorText ).title().head().lf().br().body();
  errDoc.h1("Cgipp Error. " ).br();
  errDoc.font('-',2 ).put( S ).font();
  errDoc.p().p().h2().put("Specification: " ).put( errorText ).h2();
  errDoc << br << p << p;
  errDoc << "Cgipp encountered an error (" << (int)signal
	 << ") while executing the cgi code.  If the error "
	 << "is due to low memory or an input/output " 
	 << "difficulty, please try re-calling the program."
	 << "Otherwise, contact your webmaster or system "
	 << "administrator for assistance." << br << p;
  errDoc.body().foot().hr(50).p();
  errDoc.address().put( "Cgipp by Stig E. Sand " );
  errDoc.link( "mailto:stig@ii.uib.no", "(stig@ii.uib.no)" );
  errDoc.address().foot().html();

  // Terminate as a complete failure
  exit( EXIT_FAILURE );
}

Cgipp::Cgipp(void) {
  
  // Give the process thirty seconds,  should be plenty.
#ifdef unix
  signal( SIGALRM, die );
  alarm( 30 );
#endif

#if CGIPPDEBUG
  int debugging = CGITRUE;
#else
  int debugging = CGIFALSE;
#endif
  
  if ( debugging == CGITRUE ) {
    time_t now;    tm *date;
    char s[80];
    
    now = time( NULL );
    date = localtime( &now );
#ifdef HAVE_STRFTIME    
    strftime( s, 80, "%A, %B %d, %Y %I:%M:%S %p", date );
#else
    sprintf( S, "Warning: lacking strftime()..");
#endif    
    DEBUGPUT("\nCgipp debugging log started on " <<  s);
  }

  itsFirstEntry=0;
  itsEnv = new CgiEnv();

  if (MAINT) DEBUGPUT("Creating the Cgi Object");
  if ( noCaseCompare( itsEnv->getReqMethod(), "post" ) )
    postFormInput();
  else
    getFormInput();

};


Cgipp::~Cgipp(){

  if (MAINT) DEBUGPUT("Destructing the data: ");
  if (itsFirstEntry) itsFirstEntry->dumpall();
  
  if (itsFirstEntry) delete itsFirstEntry;
  delete itsEnv;
  if (MAINT) DEBUGPUT("Cgipp-class destr. and logging ended");
  cout << endl; // To make sure stdout is flushed.
};


int
Cgipp::parseFormInput( const char *data, int length ){

  /*SDS
   * Parses all the input in <i>data</i> and stores it in the
   * list.
   */
  if (MAINT) DEBUGPUT("Parsing: " << data << "$$" << length);
  int pos = 0;
  FormEntry *n;
  FormEntry *l = NULL;

  while ( pos != length ) { 
    int foundEq = 0, foundAmp = 0, len = 0;
    int start = pos;
    char *attr, *value;
    
    while ( pos != length ) { 
      if ( data[pos] == '=' ){
	foundEq = 1;
	pos++; break;
      }
      pos++; len++;
    }
      
    if (!foundEq){break;}
      
    if ( ! unEscapeChars( &attr, data+start, len ) ) 
      _CgiError("Insuff. Mem. (Cgipp::parseFormInput)");

    start = pos;
    len = 0;
    while (pos != length) { /* pos != length */
      if (data[pos] == '&'){
	foundAmp = 1;
	pos++; break;
      }
      pos++; len++;
    }
      
    /* The last pair probably won't be followed by a &, but
       that's fine, so check for that after accepting it */
    if ( ! unEscapeChars( &value, data+start, len ) ) 
      _CgiError("Insuff. Memory (Cgipp::parseFormInput)");
      
    /* OK, we have a new pair, add it to the list. */
    if (MAINT) DEBUGPUT("Pair: |" <<attr<< "| |" <<value<< "|");
    n = new FormEntry(attr, value);	
    
    if ( ! l ) itsFirstEntry = n;
    else l->setNext(n);
      
    l = n;
    if ( ! foundAmp )	break;
  }
  
  return CGIOK;
}


FormEntry*
Cgipp::findNextEntry(const char *findTarget, FormEntry *findPos) const {
  /*SDS
   * Functions who finds the next entry from position <i>findPos</i>
   * with name <i>findTarget</i>
   */
  if (MAINT) DEBUGPUT("Searching for " << findTarget);
  FormEntry *c = findPos;
  
  while (c){
    if (!strcmp(c->Attr(), findTarget))      return c;
    c = c->Next();
  }
  
  return NULL;
};

char* 
Cgipp::getvar( char *name, int max, int& found, int newlines/*=1*/ ) {
  /*SDS
   * Fetches the variable named <i>name</i>
   */

  if (!name) return 0L;
  
  FormEntry *e = findEntry( name );
  
  if ( ! e ){
    found = CGIFALSE;
    return 0L;
  }
  
  if (!e->Value()) {
    found = CGITRUE;
    return NULL;
  }
  char *temp = new char[strlen(e->Value())+1];

  if ( ! temp ) _CgiError( "Insuff. Memory (Cgipp::getvar)" );

  if ( ! makeString( e, temp, max, newlines ) ) {
    found = CGITRUE;
    return 0L;
  }
  found = CGITRUE;
  
  return temp;
};

int 
Cgipp::getint( char *name, int& found ){
  /*SDS
   * Fetches an integer variable named <i>name</i>
   * The method uses atoi() in stdlib.h
   * Returns 0 if something wrong happened. (if that causes problems,
   * fetch it as a string and handle it yourself)
   */

  FormEntry *e = findEntry( name );  
  if ( ! e || ! strlen( e->Value() ) ) {
    found = CGIFALSE;
    return 0;
  }	
  int ch = firstNonspaceChar( e->Value() );
  if ( ! ( isdigit( ch ) ) && ( ch != '-' ) && ( ch != '+' ) ) {
    found = CGIFALSE;
    return 0;
  } 
  else {
    found = CGITRUE;
    int temp = atoi( e->Value() );
    return temp;
  }
};

double
Cgipp::getdouble( char *name, int& found ){
  /*SDS
   * Fetches a double variable named <i>name</i>
   * The method uses atof() in stdlib.h
   * Returns 0 if something wrong happened. (if that causes problems,
   * fetch it as a string and handle it yourself)
   */
  
  FormEntry *e = findEntry( name );
  if ( ! e || ! strlen( e->Value() ) ) {
    found = CGIFALSE;
    return 0;
  } 
  int ch = firstNonspaceChar( e->Value() );
  if ( ! isdigit( ch) && ch != '.' && ch != '-' && ch != '+' ) {
    found = CGIFALSE;
    return 0;
  } 
  else {
    found = CGITRUE;
    double temp = atof( e->Value() );
    return temp;
  }
}

char** 
Cgipp::getmultivar( char *name, int& numFound ){
  /*SDS
   * Fetches several strings with the same variable name
   */
  
  int index = 0;
  char **result = 0;
  
  FormEntry *e = findEntry( name );
  
  if ( ! e ) { numFound = 0; return NULL;  }
  
  do {

    if ( ! result ){
      result = new char*[index + 1];
      if ( ! result )_CgiError("Insuff. Mem. (Cgipp::getmultivar");
    }
    else {	
      char** temp = new char*[ index + 1 ];       // get new space
      
      if(!temp) _CgiError("Insuff. Mem. (Cgipp::getmultivar" );
      
      // copy old into new, minus next element
      for ( int i = 0; i < index; i++ ){
	temp[i] = new char[ strlen( result[i] ) + 1 ];
	strcpy( temp[i], result[i] );
      }
      
      if (result) delete [] result;           // delete old space
      result = temp;                          // install new
      
    }
    
    result[index] = new char[ strlen(e->Value()) + 1 ];
    if ( ! result[index] )
      _CgiError("Insuff Mem (Cgipp::getmultivar." );
    
    strcpy( result[index], e->Value() );
    index++;

    e=findNextEntry(name, e->Next());

  }
  while (e);
  
  numFound = index;
  return result;
}

int 
Cgipp::makeString( FormEntry *e, char *result, int max, int newlines) const {
  /*SDS
   * Creates a string to fit the FormEntry.
   */
  
  char *dp;
  int truncated = 0, len = 0, avail = max-1;
  int crCount = 0, lfCount = 0;	
  
  dp = result;
  const char *sp = e->Value();	
  
  while (1) {
    int ch = *sp;
    if ( len == avail ) {truncated = 1;break;}	 
    /* Fix the CR/LF, LF, CR nightmare: watch for
     * consecutive bursts of CRs and LFs in whatever
     * pattern, then actually output the larger number
     * of LFs. Consistently sane, yet it still allows
     * consecutive blank lines when the user actually
     * intends them.
     */
    if ( ( ch == 13 ) || ( ch == 10 ) ) 
      if ( ch == 13 )  crCount++;
      else lfCount++;
    else {
      if ( crCount || lfCount ) {
	int lfsAdd = crCount;
	if ( lfCount > crCount ) 
	  lfsAdd = lfCount; 
	/* Stomp all newlines if desired */
	if ( ! newlines ) lfsAdd = 0;
	while ( lfsAdd ) {
	  if ( len == avail ){ truncated = 1; break; }
	  *dp = 10;
	  dp++; lfsAdd--; len++;
	}
	
	crCount = lfCount = 0;
      }
      if ( ch == '\0' ){
	/* The end of the source string */
	break;				
      }	
      *dp = ch;
      dp++; len++;
    }
    sp++;	
  }	
  *dp = '\0';
  
  if ( truncated ) return -1;
  else if (!len) return 0;
  else return 1;
};


int 
Cgipp::unEscapeChars( char **sp, const char *cp, int len ) const {
  /*SDS
   * Unescapes the chars in sp
   */

  enum EscapeType { Rest = 0, First, Second };
      
  char 		hexValues[ 256 ];
  char 		*s;
  EscapeType 	escapeState = Rest;
  int 		escapedValue=0, srcPos=0, dstPos = 0;

  int i;
  for ( i = 0; i < 256; i++ ) hexValues[i] = 0;
 
  hexValues['0'] = 0;	  hexValues['1'] = 1;	
  hexValues['2'] = 2;	  hexValues['3'] = 3;	
  hexValues['4'] = 4;	  hexValues['5'] = 5;	
  hexValues['6'] = 6;	  hexValues['7'] = 7;	
  hexValues['8'] = 8;	  hexValues['9'] = 9;
  hexValues['A'] = 10;    hexValues['B'] = 11;
  hexValues['C'] = 12;    hexValues['D'] = 13;
  hexValues['E'] = 14;    hexValues['F'] = 15;
  hexValues['a'] = 10;    hexValues['b'] = 11;
  hexValues['c'] = 12;    hexValues['d'] = 13;
  hexValues['e'] = 14;    hexValues['f'] = 15;

  s = new char[ len + 1 ];
  
  if ( ! s ) _CgiError("Insuff Mem (unescapeChars)");
  
  while ( srcPos < len ) {
    int ch = cp[ srcPos ];
    
    switch ( escapeState ){
    case Rest:
      if ( ch == '%' ) escapeState = First;
      else if ( ch == '+' ) s[ dstPos++ ] = ' ';
      else s[ dstPos++ ] = ch;	
      break;
    case First:
      escapedValue = hexValues[ ch ] << 4;	
      escapeState = Second;
      break;
    case Second:
      escapedValue += hexValues[ ch ];
      s[ dstPos++ ] = escapedValue;
      escapeState = Rest;
      break;
    }
    srcPos++;
  }
  s[ dstPos ] = '\0';
  *sp = s;
  return CGIOK;
};


int 
Cgipp::firstNonspaceChar( const char *s ) const {
  /*SDS
   * Returns the first nonspace char from <i>s</i>
   */
  
  int len = strspn( s, " \n\r\t" );
  return s[len];
};

